#include <assert.h>
#include <iostream.h>
#include "qlearn.h"

// ------------------------- external definitions -------------------------
extern int actionCount;
extern double collectReward(int);
extern int choose_random_int_value(int);

// ------------------------- nonmember functions --------------------------

// -------------------------- member functions ----------------------------

// destructor
Qlearn::~Qlearn(void)
{
  if (stateCount > 0) for (int i=0;i<stateCount;i++) delete [] qvalues[i];
  if (actionCount > 0) delete [] qvalues;
}

// constructor
Qlearn::Qlearn(void)
{
  minState = 0; maxState = 9;
  stateCount = maxState - minState + 1;
  resetState();

  qvalues = new double * [actionCount]; assert(qvalues);
  for (int i=0;i<actionCount;i++) {
    qvalues[i] = new double [stateCount]; assert(qvalues[i]);
    for (int j=0;j<stateCount;j++) qvalues[i][j] = 0.0;
  }

  learningRate = 0.9;
  discountFactor = 0.95;
}

void
Qlearn::resetState(void) 
{
  currentState = choose_random_int_value(stateCount - 1);
}

void
Qlearn::performAction(Action newAction)
{

  // what is the state after the action
  int newState = evaluateAction(newAction);

  // what is the reward associated with the new state
  double reward = collectReward(newState);

  // what is the maximum expectation of future reward
  double expectation = findMaxExpectation(newState);
  expectation *= discountFactor;
  
  double qvalue =  qvalues[newAction][currentState];
  double difference = reward + expectation - qvalue;

  qvalues[newAction][currentState] = qvalue + learningRate*difference;

  cerr << "oldState: " << currentState << " action: " << newAction;
  cerr << " newState: " << newState << " reward: " << reward << endl;

  currentState = newState;
}

int
Qlearn::evaluateAction(Action newAction) 
{
  int newState;

  // find the new state
  switch (newAction) {
  case advance:  newState = currentState - 1; break;
  case hesitate: newState = currentState; break;
  case retreat:  newState = currentState + 1; break;
  default: 
    cerr << "Qlearn::evaluateAction() -- unknown action: ";
    cerr << newAction << endl;
  }

  // make sure that the new state is not outside designated limits
  if (newState < minState) newState = minState;
  else if (newState > maxState) newState = maxState;
  
  return newState;
}

double
Qlearn::findMaxExpectation(int state)
{
  double maxExpectation = qvalues[0][state];

  for (int i=1;i<actionCount;i++) {
    double newExpectation = qvalues[i][state];
    if (newExpectation > maxExpectation) maxExpectation = newExpectation;
  }

  return maxExpectation;
}

// find the best action to take from the current state
Action
Qlearn::chooseBestAction(void)
{
  double maxExpectation = qvalues[0][currentState];
  Action bestAction = (Action) 0;

  for (int i=1;i<actionCount;i++) {
    double newExpectation = qvalues[i][currentState];
    if (newExpectation > maxExpectation) {
      maxExpectation = newExpectation;
      bestAction = (Action) i;
    }
  }

  return bestAction;
}
